import WebSocket, { WebSocketServer } from 'ws';
import { createClient } from 'redis';
import { exec } from 'child_process';

// dotenv config
import * as dotenv from 'dotenv';
dotenv.config();


const wss = new WebSocketServer({ port: process.env.PORT || 8080 });
const workers = [];
const clients = [];
const queue = [];

const alphabet = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';

const redis = createClient({
    socket: {
        host: process.env.REDIS_HOST,
    }
});

redis.on('error', err => console.log('Redis Client Error', err));
redis.on('connect', () => console.log('Redis Client Connected'));

await redis.connect();

setInterval(() => {
    const freeWorkers = workers.filter((worker) => !worker.hash);
    freeWorkers.forEach((worker) => {
        worker.send('exit');
        worker.close();
    });
    workers.splice(0, freeWorkers.length);
}, 60_000);

wss.on('connection', (ws) => {
    ws.on('message', handleFirstMessage);
    ws.on('close', () => {
        if(ws.isWorker) {
            console.log('Worker disconnected');
            workers.splice(workers.indexOf(ws), 1);
            return;
        }
        console.log('Client disconnected');
        clients.splice(clients.indexOf(ws), 1);
    });
});

wss.on('open', () => console.log(`WebSocket server started on port ${process.env.PORT || 8080}`));

function handleWorkerMessage(message) {
    console.log(`Received worker message => ${message}`);
    if(!message.startsWith('found')) return;
    const [_, hash, password] = message.split(' ');
    thisHashWorkers = workers.filter((worker) => worker.hash === hash);
    thisHashWorkers.forEach((worker) => {
        worker.send(`stop`)
        worker.hash = null;
    });
    redis.set(hash, password);
    clients.forEach((client) => client.send(`found ${hash} ${password}`));
}

async function handleClientMessage(buffer) {
    const message = buffer.toString();
    const [hash, minLen, maxLen, workersCount] = message.split(':');
    console.log('Received client message:');
    console.table({ hash, minLen, maxLen, workersCount });
    const alreadyFound = await redis.get(hash);
    console.log('hash in redis: ', alreadyFound ? 'yes' : 'no')
    if(alreadyFound) return this.send(`found ${hash} ${alreadyFound}`);
    const freeWorkers = workers.filter((worker) => !worker.hash);
    console.log('free workers: ', freeWorkers.length);
    const missingWorkers = workersCount - freeWorkers.length;
    if(missingWorkers > 0) {
        console.log(`missing workers: ${missingWorkers}, scaling up...`);
        exec(`docker-compose up -d --scale worker=${workers.length + missingWorkers}`);
    }
    console.log(`Adding to queue`);
    queue.push({ hash, needed: missingWorkers, count: workersCount, minLen, maxLen });
}

function base10ToBase62(number) {
    let result = '';
    while(number > 0) {
        result = alphabet[number % alphabet.length] + result;
        number = Math.floor(number / alphabet.length);
    }
    return result;
} 

function handleFirstMessage(message) {
    console.log(`Received first message => ${message}`);
    this.removeListener('message', handleFirstMessage);
    if(message !== 'slave') {
        this.on('message', handleClientMessage);
        clients.push(this);
        return;
    }
    workers.push(this);
    this.on('message', handleWorkerMessage);
    console.log('Queue', queue);
    if(queue.length) {
        const { hash, needed, count, minLen, maxLen } = queue[0];
        const start = alphabet.length ** minLen;
        const end = alphabet.length ** maxLen;
        const interval = (end - start) / count;
        const currentIndex = count - needed + i;
        const currentStart = start + (interval * currentIndex);
        const currentEnd = start + (interval * (currentIndex + 1));
        this.hash = hash;
        console.log(`Sending search command to worker`);
        console.table({ hash, currentStart, currentEnd });
        this.send(`search ${hash} ${base10ToBase62(currentStart)} ${base10ToBase62(currentEnd)}`);
        queue[0] = {...queue[0], needed: needed - 1};
    }
    
}